from argparse import ArgumentParser
from sys import argv
from .oauth_flow import delete_access_token, run_oauth_flow, verify_access_token
from .repo_keys import list_repo_keys


def auth_entry():
    parser = ArgumentParser(prog='auth', description='Manages authorization to GitHub')
    auth_choice = 'auth'
    verify_choice = 'verify'
    delete_choice = 'delete'
    parser.add_argument('action', choices=[auth_choice, verify_choice, delete_choice],
                        help='Get authorization, verify existing authorization, ' +
                             'or remove your local authorization',
                        nargs='?',
                        default=auth_choice)
    args = parser.parse_args(argv[1:])

    actions = {
        auth_choice: run_oauth_flow,
        verify_choice: verify_access_token,
        delete_choice: delete_access_token,
    }

    command = actions[args.action]
    command()


def list_keys_entry():
    parser = ArgumentParser(prog='listkeys',
                            description='Lists the deploy keys for the given repro')
    parser.add_argument('repo_name', help='The full name of the repo (e.g., octocat/Hello-World)')
    args = parser.parse_args(argv[1:])
    repo_name = args.repo_name

    if repo_name.startswith('\'') or repo_name.startswith('"'):
        repo_name = repo_name[1:]

    if repo_name.endswith('\'') or repo_name.endswith('"'):
        repo_name = repo_name[:-1]

    list_repo_keys(repo_name)



from fabric.connection import Connection
from invoke.watchers import Responder

from .permissions import _exists, ensure_directory

DEFAULT_KEY_DIRECTORY = '/var/deploykeys'


def get_keyfile(project_name, directory=DEFAULT_KEY_DIRECTORY, public=True):
    sanitized_name = project_name.replace('/', '_')
    suffix = '.pub' if public else ''
    keyfile = f'{directory}/{sanitized_name}{suffix}'
    return keyfile


def create_key(conn: Connection, project_name, owning_group, path=DEFAULT_KEY_DIRECTORY):
    ensure_directory(conn, path, owning_group, mod='600')
    keyfile = get_keyfile(project_name, path, public=False)
    if _exists(conn, keyfile, sudo=True):
        conn.sudo(f'rm {keyfile}')
    # Automatically provide an empty passphrase
    responder = Responder(r'passphrase.*:', '\n')
    conn.sudo(f'ssh-keygen -t rsa -b 4096 -C "{project_name}" -f {keyfile}',
              watchers=[responder])



from typing import Optional

from fabric.connection import Connection
from github import GithubException
from .oauth_flow import get_api
from .fabric_commands import get_keyfile


def list_repo_keys(repo_full_name):
    api = get_api()

    repo = api.get_repo(repo_full_name)
    deploy_keys = repo.get_keys()

    try:
        for key in deploy_keys:
            print(key.title)
    except GithubException as exception:
        print(f'Failed with {exception.status}, data: {exception.data}')


def add_repo_key(conn: Connection, repo_full_name: str, key_name: Optional[str] = None):
    public_keyfile = get_keyfile(repo_full_name)
    public_keyfile_contents = conn.sudo(f'cat {public_keyfile}').stdout
    public_keyfile_contents = public_keyfile_contents.rstrip()

    if not key_name:
        # Default the key_name to the hostname we are connected to
        key_name = conn.run('hostname').stdout
        key_name = key_name.rstrip()

    api = get_api()

    repo = api.get_repo(repo_full_name)

    try:
        print(f'key_name: "{key_name}"')
        repo.create_key(key_name, public_keyfile_contents, read_only=True)
    except GithubException as exception:
        print(f'Failed with {exception.status}, data: {exception.data}')
        raise exception



from typing import Iterable

from colorama import init, Fore, Style
from colorama.initialise import orig_stdout
from fabric.connection import Connection
from invoke import UnexpectedExit
from invoke.watchers import Responder

from .permissions import _exists
from .ssh_key import get_keyfile


def _ensure_colorama_init():
    if not orig_stdout:
        init()


def prepare_user(conn: Connection, user: str, group: str, add_sudo=True, no_sudo_passwd=False):
    messages = ''
    user_exists = False

    try:
        conn.sudo(f'getent passwd {user}')
        user_exists = True
    except UnexpectedExit:
        pass

    if not user_exists:
        # Handle all the prompts for information about the new user like their name and room number
        responder = Responder(r'.*\[.*\].*', '\n')
        conn.sudo(f'adduser --disabled-password {user}', pty=True, watchers=[responder])

    group_exists = False
    try:
        conn.sudo(f'getent group {group}')
        group_exists = True
    except UnexpectedExit:
        pass

    if not group_exists:
        conn.sudo(f'addgroup {group}')

    conn.sudo(f'adduser {user} {group}')

    if add_sudo:
        conn.sudo(f'adduser {user} sudo')

    if no_sudo_passwd:
        sudoers_file = f'/etc/sudoers.d/{user}-plush'
        if _exists(conn, sudoers_file, sudo=True):
            conn.sudo(f'rm {sudoers_file}')
        conn.sudo(f"echo '{user} ALL=(ALL:ALL) NOPASSWD:ALL' | sudo EDITOR='tee -a' visudo -f " +
                  f"{sudoers_file}", pty=True)

    if add_sudo and not no_sudo_passwd:
        messages += f'Ensure {user} has a secure password configured ' \
                     'in order to run commands as sudo.\n' \
                     'Alternatively, rerun this command with ' \
                     'the no_sudo_passwd=True parameter\n'
    if not user_exists:
        messages += f'The {user} account was created without a password set. ' \
                     'If a password is needed, a superuser will need to manually run: \n' \
                    f'sudo passwd {user}\n' \
                     'Where possible, consider not setting a password and ' \
                     'instead using key based authentication.\n'

    if messages:
        # Every message line ends in a newline.
        # If we've got a non-empty message, strip the last character
        messages = messages[:-1]

    return messages


def add_authorized_key(conn: Connection, user, public_key):
    conn.sudo(f'mkdir -p /home/{user}/.ssh', user=user)
    conn.sudo(f'touch /home/{user}/.ssh/authorized_keys', user=user)
    conn.sudo(f'chmod -R go= /home/{user}/.ssh'.format(user), user=user)
    conn.sudo(f'echo "{public_key}" | sudo tee -a /home/{user}/.ssh/authorized_keys',
              pty=True)


def install_packages(conn: Connection, packages: Iterable[str]):
    _ensure_colorama_init()
    apt = "DEBIAN_FRONTEND=noninteractive apt-get install -y {}"
    for package in packages:
        print(Fore.GREEN + f'Installing {package}' + Style.RESET_ALL)
        conn.sudo(apt.format(package))



from fabric.connection import Connection
from ..fabric_commands import install_packages
from .ssh_key import get_keyfile


def clone(conn: Connection, repo_full_name: str, path: str, skip_strict_key_checking=False):
    install_packages(conn, ['git'])
    keyfile = get_keyfile(repo_full_name, public=False)
    conn.sudo('git config --system protocol.ext.allow always')
    strict_key_checking = '-o StrictHostKeyChecking=no ' if skip_strict_key_checking else ''
    conn.sudo(f"git clone 'ext::ssh -i {keyfile} {strict_key_checking}git@github.com %S " \
              f"{repo_full_name}.git' {path}")



from typing import Optional
from fabric.connection import Connection
from patchwork.files import exists as patchwork_exists


def _exists(conn: Connection, path: str, sudo: bool=False) -> bool:
    # pylint doesn't understand the @set_runner decorator
    # create a wrapper so we only have to suppress the error once
    return patchwork_exists(conn, path, sudo=sudo) # pylint: disable=E1120


def ensure_directory(conn: Connection,
                     directory: str,
                     owning_group: Optional[str] = None,
                     owning_user: Optional[str] = None,
                     mod: str = 'ug+rwX,o+rX,o-w'):
    conn.sudo(f'mkdir -p {directory}')
    set_permissions_directory(conn, directory, group=owning_group, user=owning_user, mod=mod)


def set_permissions_directory(conn: Connection, # pylint: disable=R0913
                              directory: str,
                              group: Optional[str] = None,
                              user: Optional[str] = None,
                              mod: str = '660',
                              setgid: bool = True):
    if group is not None:
        conn.sudo(f'chgrp -R {group} {directory}')

    if user is not None:
        conn.sudo(f'chown -R {user} {directory}')

    conn.sudo(f'chmod -R {mod} {directory}')

    if setgid:
        conn.sudo(f'chmod -R g+s {directory}')


def set_permissions_file(conn: Connection,
                         file: str,
                         user: Optional[str] = None,
                         group: Optional[str] = None,
                         mod: str = '644'):
    if group is not None:
        conn.sudo(f'chgrp {group} {file}')

    if user is not None:
        conn.sudo(f'chown {user} {file}')

    conn.sudo(f'chmod {mod} {file}')
